from ..settings import Settings
from ..app_config import AppConfig
from ..lexemes.lexemes import Lexemes
from ..helpers.theme_helper import ThemeHelper
from ..helpers.clipboard_helper import ClipboardHelper
from ..helpers.tooltip_helper import TooltipHelper



"""
Notolog Editor
Open-source markdown editor developed in Python.

File Details:
- Purpose: Extends AI Assistant Dialog Class by formatting and displaying messages from and to the user.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtWidgets import QStyle, QLabel, QPushButton
from PySide6.QtGui import QTextDocument, QColor
from PySide6.QtCore import Qt

from . import AppConfig
from . import Lexemes
from . import ThemeHelper
from . import ClipboardHelper
from . import TooltipHelper

import logging


class AIMessageLabel(QLabel):

    button: QPushButton = None

    def __init__(self, parent=None, text=None, settings=None):
        super(AIMessageLabel, self).__init__(parent)

        self.parent = parent

        if self.parent and hasattr(self.parent, 'font'):
            # Apply font from the main window to the element
            self.setFont(self.parent.font())

        # The class may be in use within settings itself
        self.settings = settings if settings else self.parent.settings  # type: ignore

        self.logger = logging.getLogger('ai_message_label')

        self.logging = AppConfig().get_logging()
        self.debug = AppConfig().get_debug()

        # Load lexemes for selected language and scope
        self.lexemes = Lexemes(self.settings.app_language, default_scope='ai_assistant')

        self.theme_helper = ThemeHelper()

        # Set text
        if text:
            self.setText(text)

        self.init_ui()

    def init_ui(self):

        self.button = QPushButton(self)

        # copy_icon = self.style().standardIcon(QStyle.StandardPixmap.SP_DirIcon)  # Set an icon using standard icons
        copy_icon = self.theme_helper.get_icon(
            theme_icon='copy.svg',
            color=QColor(self.theme_helper.get_color('ai_assistant_message_icon_color', css_format=True))
        )
        self.button.setIcon(copy_icon)

        self.button.clicked.connect(self.copy_content)
        self.button.setCursor(Qt.CursorShape.ArrowCursor)
        self.button.setFixedSize(20, 20)
        self.button.setStyleSheet("QPushButton { border: none; padding: 0; }")
        self.button.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        # Ensure the text doesn't overlap the button
        button_size = self.button.sizeHint()
        self.setContentsMargins(0, 0, button_size.width(), 0)

    def resizeEvent(self, event):
        button_size = self.button.sizeHint()
        frame_width = self.style().pixelMetric(QStyle.PixelMetric.PM_DefaultFrameWidth)
        self.button.move(self.rect().right() - frame_width - button_size.width(),
                         (self.rect().top() + button_size.height() + 2) // 2)
        super(AIMessageLabel, self).resizeEvent(event)

    def copy_content(self):
        # Copy text to the clipboard
        ClipboardHelper.set_text(self.get_text_to_copy())
        # Show tooltip
        TooltipHelper.show_tooltip(widget=self.button, text=self.lexemes.get('dialog_message_copied_tooltip'))

    def get_text_to_copy(self):
        # Check if the result needs to be converted from HTML to raw text.
        if getattr(self.settings, 'ai_config_convert_to_md', False):
            # Create a QTextDocument from the label's HTML content.
            doc = QTextDocument()
            doc.setHtml(self.text())

            # Convert the HTML content to raw text.
            return doc.toRawText()

        return self.text()



# tests/test_lexemes.py

from notolog.lexemes.lexemes import Lexemes

from logging import Logger

import os
import pytest
import pathlib
import importlib.util


class TestLexemes:

    @pytest.fixture(scope="function")
    def test_obj_lexemes(self, mocker, request):
        """
        Testing object fixture.
        May contain minimal logic for testing purposes.
        """

        # Retrieve parameter values from the test request.
        language, lexemes, lexemes_dir = request.param

        if lexemes:
            """
            If lexemes explicitly set return them.
            This is for test cases where the lexemes file read is not necessary.
            """
            mocker.patch.object(Lexemes, 'load_lexemes', return_value=lexemes)
        elif lexemes_dir:
            """
            If lexemes dir is set read lexemes from there.
            For test cases to check actual lexemes file read.
            """
            test_lexemes_dir = os.path.join(os.path.dirname(__file__), lexemes_dir)
            mocker.patch.object(Lexemes, 'get_lexemes_dir', return_value=test_lexemes_dir)
        else:
            """
            Return empty lexemes in the rest of the cases.
            """
            mocker.patch.object(Lexemes, 'load_lexemes', return_value={})

        yield Lexemes(language=language)

    @pytest.fixture(scope="function")
    def test_exp_params_fixture(self, request):
        """
        Fixture to pass params and get expected results.
        This method is used only for passing params via pytest fixture.
        """

        # Retrieve parameter values from the test request.
        param_values = request.param

        yield param_values

    @pytest.mark.parametrize(
        "test_obj_lexemes, test_exp_params_fixture",
        [
            ((None, None, None), ('en', {})),
            (('', {}, None), ('en', {})),
            (('en', {123: 456}, None), ('en', {123: 456})),
            (('anyval', {'A': 'B', 'C': 'D'}, None), ('anyval', {'A': 'B', 'C': 'D'})),
        ],
        indirect=True
    )
    def test_lexemes_init(self, test_obj_lexemes: Lexemes, test_exp_params_fixture):
        """
        Test and check initial params are exist in the testing object.
        """
        assert isinstance(test_obj_lexemes.logger, Logger)

        exp_language, exp_lexemes = test_exp_params_fixture

        assert hasattr(test_obj_lexemes, 'language')
        assert test_obj_lexemes.language == exp_language
        assert hasattr(test_obj_lexemes, 'lexemes')
        assert test_obj_lexemes.lexemes == exp_lexemes

    @pytest.mark.parametrize(
        "test_obj_lexemes, test_exp_params_fixture",
        [
            ((None, None, None), (None, 'en')),
            (('', None, None), (None, 'en')),
            (('en', None, None), (None, 'en')),
            (('anyval', None, None), (None, 'anyval')),
            (('anyval', None, None), ('new', 'new')),
        ],
        indirect=True
    )
    def test_lexemes_set_language(self, test_obj_lexemes: Lexemes, test_exp_params_fixture):
        """
        Test language setting for lexemes after the one was initialized.
        """
        param_language, exp_language = test_exp_params_fixture

        if param_language:
            test_obj_lexemes.set_language(param_language)

        assert hasattr(test_obj_lexemes, 'language')
        assert test_obj_lexemes.language == exp_language

    @pytest.mark.parametrize(
        "test_obj_lexemes, test_exp_params_fixture",
        [
            ((None, None, None), 'en'),
            (('', None, None), 'en'),
            (('en', None, None), 'en'),
            (('any', None, None), 'en'),
            (('..', None, None), 'en'),
        ],
        indirect=True
    )
    def test_lexemes_get_lexemes_dir(self, mocker, test_obj_lexemes: Lexemes, test_exp_params_fixture):
        """
        Test lexemes target dir path.
        """
        exp_lexemes_dir = test_exp_params_fixture

        mocker.patch.object(os.path, 'dirname', return_value='')

        assert test_obj_lexemes.get_lexemes_dir() == exp_lexemes_dir

    @pytest.mark.parametrize(
        "test_obj_lexemes, test_exp_params_fixture",
        [
            ((None, None, None),
             ('en', {})),
            ((None, None, 'test_lexemes'),
             ('en', {'test_common': {'any': 'thing', 'hello': 'Hello World!', 'test': 'Passed'},
                     'test_scope': {123: 456, 'foo': 'bar', 'test': 'result'}})),
        ],
        indirect=True
    )
    def test_lexemes_load_lexemes(self, test_obj_lexemes: Lexemes, test_exp_params_fixture):
        """
        Test actual lexemes loaded.
        """
        exp_language, exp_lexemes = test_exp_params_fixture

        assert hasattr(test_obj_lexemes, 'lexemes')
        assert test_obj_lexemes.lexemes == exp_lexemes

    @pytest.mark.parametrize(
        "test_obj_lexemes, test_exp_params_fixture",
        [
            ((None, None, None), (None, None, None)),
            ((None, None, 'test_lexemes'), ('test_common', 'hello', 'Hello World!')),
            ((None, None, 'test_lexemes'), ('test_common', 'test', 'Passed')),
            ((None, None, 'test_lexemes'), ('test_scope', 123, 456)),
            ((None, None, 'test_lexemes'), ('test_scope', 'undefined_name', '')),
            ((None, None, 'test_lexemes'), ('test_scope_undefined', None, None)),
            ((None, None, 'test_lexemes'), ('test_scope_undefined', 'undefined_name', None)),
        ],
        indirect=True
    )
    def test_lexemes_get_all(self, test_obj_lexemes: Lexemes, test_exp_params_fixture):
        """
        Test getter for correctly loaded and really existent lexemes.
        """
        param_scope, param_name, exp_result = test_exp_params_fixture

        _all_lexemes = test_obj_lexemes.get_all()

        if param_scope:
            assert len(_all_lexemes) > 0
            _scope_lexemes = _all_lexemes[param_scope] if param_scope in _all_lexemes else None
            if _scope_lexemes:
                if param_name in _scope_lexemes:
                    assert _scope_lexemes[param_name] == exp_result
                else:
                    assert exp_result == ''
            else:
                assert exp_result is None

    @pytest.mark.parametrize(
        "test_obj_lexemes, test_exp_params_fixture",
        [
            ((None, None, None), (None, None, None)),
            ((None, None, 'test_lexemes'), ('test_common', 'hello', 'Hello World!')),
            ((None, None, 'test_lexemes'), ('test_common', 'test', 'Passed')),
            ((None, None, 'test_lexemes'), ('test_scope', 123, 456)),
            ((None, None, 'test_lexemes'), ('test_scope', 'undefined_name', '')),
            ((None, None, 'test_lexemes'), ('test_scope_undefined', None, None)),
            ((None, None, 'test_lexemes'), ('test_scope_undefined', 'undefined_name', None)),
        ],
        indirect=True
    )
    def test_lexemes_get_by_scope(self, test_obj_lexemes: Lexemes, test_exp_params_fixture):
        """
        Test getter for correctly loaded and really existent lexemes.
        """
        param_scope, param_name, exp_result = test_exp_params_fixture

        _scope_lexemes = test_obj_lexemes.get_by_scope(param_scope)
        if not _scope_lexemes:
            assert exp_result is None
        else:
            if param_name in _scope_lexemes:
                assert _scope_lexemes[param_name] == exp_result
            else:
                assert exp_result == ''

    @pytest.mark.parametrize(
        "test_obj_lexemes, test_exp_params_fixture",
        [
            ((None, None, None), (None, None, None)),
            ((None, None, 'test_lexemes'), ('test_common', 'hello', 'Hello World!')),
            ((None, None, 'test_lexemes'), ('test_common', 'test', 'Passed')),
            ((None, None, 'test_lexemes'), ('test_scope', 123, 456)),
            ((None, None, 'test_lexemes'), ('test_scope', 'undefined_name', '')),
            ((None, None, 'test_lexemes'), ('test_scope_undefined', None, None)),
            ((None, None, 'test_lexemes'), ('test_scope_undefined', 'undefined_name', None)),
        ],
        indirect=True
    )
    def test_lexemes_get(self, test_obj_lexemes: Lexemes, test_exp_params_fixture):
        """
        Test getter for correctly loaded and really existent lexemes.
        """
        param_scope, param_name, exp_result = test_exp_params_fixture

        assert test_obj_lexemes.get(param_name, param_scope) == exp_result

    @staticmethod
    def load_module_from_file(full_path_to_module, module_name):
        spec = importlib.util.spec_from_file_location(module_name, full_path_to_module)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def test_translation_files(self):
        # Define the root path for translation directories
        translations_root = os.path.join(
            pathlib.Path(os.path.dirname(__file__)).parent.resolve(), 'notolog', 'lexemes')

        # Iterate over each language directory
        for lang in os.listdir(translations_root):
            lang_path = os.path.join(translations_root, lang)
            if os.path.isdir(lang_path):
                # Iterate over each module in the language directory
                for file in os.listdir(lang_path):
                    if file.endswith('.py'):
                        # Load the module
                        module_path = os.path.join(lang_path, file)
                        module_name = f"{lang}.{file[:-3]}"  # Remove '.py' from filename, to get smth like 'toolbar'
                        module = self.load_module_from_file(module_path, module_name)

                        # You can now use 'module' to access the loaded module's attributes
                        # and ensure they are covered by some form of assertion or operation.
                        # For this example, let's just print out the module's dir(): print(dir(module))
                        assert module_name == (f"{os.path.basename(lang_path)}."
                                               f"{os.path.basename(file).replace('.py', '')}")

                        # Check the actual language data exists
                        assert isinstance(module.lexemes, dict)



from notolog.enums.enum_base import EnumBase

assert hasattr(EnumBase, 'default'), "Check default() method is implemented in the base class"


class SomethingEnum(EnumBase):
    """
    Enum for testing 'something' behavior or options.
    """

    SMTH1 = "Smth1"
    SMTH2 = ("Smth2", True)  # Fallback default option
    SMTH3 = "Smth3"



"""
Notolog Editor
Open-source markdown editor developed in Python.

File Details:
- Purpose: Part of the base module.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from ..settings import Settings
from ..app_config import AppConfig
from ..lexemes.lexemes import Lexemes



"""
Notolog Editor
An open-source markdown editor developed in Python.

File Details:
- Purpose: Enhances QLineEdit functionality by adding an icon that triggers a QFileDialog action for dir selection.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QStyle, QLineEdit, QPushButton, QFileDialog
from PySide6.QtGui import QColor

from . import AppConfig
from . import Lexemes
from . import ThemeHelper

import logging


class DirPathLineEdit(QLineEdit):

    button: QPushButton = None

    def __init__(self, parent=None, settings=None, default_directory=None):
        super(DirPathLineEdit, self).__init__(parent)

        self.parent = parent

        if self.parent and hasattr(self.parent, 'font'):
            # Apply font from the main window to the element
            self.setFont(self.parent.font())

        # The class may be in use within settings itself
        self.settings = settings if settings else self.parent.settings  # type: ignore

        self.logger = logging.getLogger('dir_path_line_edit')

        self.logging = AppConfig().get_logging()
        self.debug = AppConfig().get_debug()

        # Load lexemes for selected language and scope
        self.lexemes = Lexemes(self.settings.app_language, default_scope='common')

        self.theme_helper = ThemeHelper()

        self.default_directory = default_directory

        self.init_ui()

    def init_ui(self):

        self.button = QPushButton(self)

        # Load the dialog icon
        self.load_icon()

        self.button.clicked.connect(self.select_dir_dialog)
        self.button.setCursor(Qt.CursorShape.ArrowCursor)
        self.button.setStyleSheet("QPushButton { border: none; padding: 0; }")
        self.button.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        # Set the default directory text if provided
        if self.default_directory:
            self.setText(self.default_directory)

        # Ensure the text doesn't overlap the button
        self.setTextMargins(0, 0, 30, 0)

    def load_icon(self):
        # Use either a standard or a custom icon for file selection.
        # folder_icon = self.style().standardIcon(QStyle.StandardPixmap.SP_DirIcon)  # Set an icon using standard icons
        icon_color = self.theme_helper.get_color('settings_dialog_hint_icon_color')
        folder_icon = self.theme_helper.get_icon(theme_icon='folder.svg', color=QColor(icon_color))
        self.button.setIcon(folder_icon)

    def resizeEvent(self, event):
        button_size = self.button.sizeHint()
        frame_width = self.style().pixelMetric(QStyle.PixelMetric.PM_DefaultFrameWidth)
        self.button.move(self.rect().right() - frame_width - button_size.width(),
                         (self.rect().bottom() - button_size.height() + 1) // 2)
        super(DirPathLineEdit, self).resizeEvent(event)

    def select_dir_dialog(self):
        directory = QFileDialog.getExistingDirectory(self, caption=self.lexemes.get('field_dir_path_dialog_caption'),
                                                     dir=self.text())
        if directory:
            self.setText(directory)



from PySide6.QtCore import QCoreApplication

from notolog.app_config import AppConfig
from notolog.modules.modules import Modules

import os
import pytest

# For headless operation
os.environ["QT_QPA_PLATFORM"] = "offscreen"


@pytest.fixture(autouse=True)
def test_core_app():

    # To correctly set up app settings
    QCoreApplication.setOrganizationName(AppConfig().get_settings_org_name())
    QCoreApplication.setApplicationName(AppConfig().get_settings_app_name_qa())

    AppConfig().set_test_mode(True)

    yield None


def is_module_available(module_name):
    try:
        result = Modules().get_by_name(module_name)
    except (ModuleNotFoundError, ImportError):
        result = False
    return result



import pytest

from .. import is_module_available


def pytest_collection_modifyitems(config, items):
    """
    Conditionally skip tests in a specific directory.
    """
    # Check if the condition is not met
    if not is_module_available('llama_cpp'):
        skip_marker = pytest.mark.skip(reason="Condition is not met. Skipping tests in 'llama_cpp'.")
        # Skip all tests in the 'llama_cpp' directory
        for item in items:
            if 'llama_cpp' in item.nodeid:
                item.add_marker(skip_marker)



# tests/ui_tests/test_search_form.py

from notolog.ui.search_form import SearchForm
from notolog.ui.toolbar import ToolBar
from notolog.notolog_editor import NotologEditor
from notolog.settings import Settings
from notolog.enums.languages import Languages
from notolog.modules.modules import Modules

from . import test_app  # noqa: F401

import pytest


class TestSearchForm:

    @pytest.fixture(scope="class", autouse=True)
    def settings_obj(self):
        """
        Use 'autouse=True' to enable automatic setup, or pass 'settings_obj' directly to main_window()
        """
        # Fixture to create and return settings instance
        settings = Settings()
        # Clear settings to be sure start over without side effects
        settings.clear()
        # Reset singleton (qa functionality)
        Settings.reload()

        yield settings

    @pytest.fixture
    def main_window(self, mocker, test_app):  # noqa: F811 redefinition of unused 'test_app'
        # Force to override system language as a default
        mocker.patch.object(Languages, 'default', return_value='la')

        # Do not show actual window; return object instance only
        mocker.patch.object(NotologEditor, 'show', return_value=None)
        # Prevent resource processing, including 'process_document_images'
        mocker.patch.object(NotologEditor, 'load_content_html', return_value=None)

        # Fixture to create and return main window instance
        yield NotologEditor(screen=test_app.screens()[0])

    @pytest.fixture(autouse=True)
    def toolbar_obj(self, main_window):
        # Fixture to create and return toolbar instance
        toolbar_obj = ToolBar(parent=main_window, actions=[], refresh=False)
        yield toolbar_obj

    @pytest.fixture(autouse=True)
    def ui_obj(self, mocker, main_window, toolbar_obj):
        # No modules
        mocker.patch.object(Modules, 'get_by_extension', return_value=[])
        # Fixture to initialize object.
        ui_obj = SearchForm(parent=toolbar_obj)
        yield ui_obj

    def test_ui_object_state(self, mocker, ui_obj: SearchForm, settings_obj):
        # Check app language set correctly
        assert settings_obj.app_language == 'la'

        # Searched text update
        assert ui_obj.text() == ''
        ui_obj.set_text('test_text')
        assert ui_obj.text() == 'test_text'

        # Check searched occurrences values
        assert ui_obj._search_pos_label.text() == ''
        assert ui_obj._search_count_label.text() == ''
        # Set counter value
        ui_obj.set_counter_text('123')
        assert ui_obj._search_pos_label.text() == ''  # remains the same
        assert ui_obj._search_count_label.text() == '123'  # updated
        # Set index position value
        ui_obj.set_position_text('111')
        assert ui_obj._search_pos_label.text() == '111'  # updated
        assert ui_obj._search_count_label.text() == '123'  # remains the same
        # Set new counter value
        ui_obj.set_counter_text('456')
        assert ui_obj._search_pos_label.text() == ''  # reset
        assert ui_obj._search_count_label.text() == '456'  # updated



# tests/ui_tests/test_settings_dialog.py

from PySide6.QtCore import Qt, QObject
from PySide6.QtWidgets import QWidget, QTabWidget, QLabel, QCheckBox, QLineEdit, QSlider
from PySide6.QtWidgets import QScrollArea, QVBoxLayout, QSizePolicy, QPlainTextEdit, QComboBox, QSpinBox, QDoubleSpinBox
from PySide6.QtTest import QTest

from notolog.ui.settings_dialog import SettingsDialog
from notolog.ui.enum_combo_box import EnumComboBox
from notolog.ui.label_with_hint import LabelWithHint
from notolog.ui.dir_path_line_edit import DirPathLineEdit
from notolog.ui.file_path_line_edit import FilePathLineEdit
from notolog.ui.horizontal_line_spacer import HorizontalLineSpacer
from notolog.notolog_editor import NotologEditor
from notolog.settings import Settings
from notolog.enums.languages import Languages
from notolog.modules.modules import Modules

from ..test_enums.something_enum import SomethingEnum

from . import test_app  # noqa: F401

import pytest

from unittest.mock import MagicMock


class TestSettingsDialog:

    @pytest.fixture(scope="class", autouse=True)
    def settings_obj(self):
        """
        Use 'autouse=True' to enable automatic setup, or pass 'settings_obj' directly to main_window()
        """
        # Fixture to create and return settings instance
        settings = Settings()
        # Clear settings to be sure start over without side effects
        settings.clear()
        # Reset singleton (qa functionality)
        Settings.reload()

        yield settings

    @pytest.fixture(autouse=True)
    def modules_obj(self):
        # Fixture to create and return modules instance
        modules = Modules()
        modules.modules = []
        yield modules

    @pytest.fixture
    def main_window(self, mocker, test_app):  # noqa: F811 redefinition of unused 'test_app'
        # Force to override system language as a default
        mocker.patch.object(Languages, 'default', return_value='la')

        # Do not show actual window; return object instance only
        mocker.patch.object(NotologEditor, 'show', return_value=None)
        # Prevent resource processing, including 'process_document_images'
        mocker.patch.object(NotologEditor, 'load_content_html', return_value=None)

        # Fixture to create and return main window instance
        yield NotologEditor(screen=test_app.screens()[0])

    @pytest.fixture(autouse=True)
    def ui_obj(self, mocker, main_window):
        # No modules
        mocker.patch.object(Modules, 'get_by_extension', return_value=[])
        # Fixture to initialize object.
        ui_obj = SettingsDialog(parent=main_window)
        yield ui_obj

    @pytest.fixture(scope="function")
    def test_params_fixture(self, request):
        # Retrieve parameter values from the test request.
        param_values = request.param

        yield param_values

    @pytest.fixture(scope="function")
    def test_exp_params_fixture(self, request):
        # Retrieve parameter values from the test request.
        param_values = request.param

        yield param_values

    def test_ui_object_state(self, mocker, ui_obj: SettingsDialog, settings_obj):
        # Check app language set correctly
        assert settings_obj.app_language == 'la'

        # Check default window title
        assert ui_obj.windowTitle() == "Configurationes"

        # Retrieve callback methods for verification.
        mock_init_ui = mocker.patch.object(ui_obj, 'init_ui', wraps=ui_obj.init_ui)

        # Check the method wasn't called as no ui_obj.exec() executed
        mock_init_ui.assert_not_called()

        # Mind the Latin translation
        assert ui_obj.tab_widget.tabText(0) == 'Generale'
        assert ui_obj.tab_widget.tabText(1) == 'Editor'
        assert ui_obj.tab_widget.tabText(2) == 'Visor'
        assert ui_obj.tab_widget.tabText(3) == 'Configuratio AI'
        assert ui_obj.tab_widget.tabText(4) == ''

        # Ensure the initial tab is the first tab
        assert ui_obj.tab_widget.currentIndex() == 0
        assert isinstance(ui_obj.tab_widget.currentWidget(), QScrollArea)
        assert ui_obj.tab_widget.currentWidget().objectName() == 'settings_dialog_tab_general'
        # Test that clicking the edit button updates the editor state
        QTest.mouseClick(
            # ui_obj.tab_widget.tabBar().tabButton(1, QTabBar.ButtonPosition.LeftSide), Qt.MouseButton.LeftButton)
            # This simply targets the middle of the tab, ensuring the interaction is as expected.
            ui_obj.tab_widget.tabBar(), Qt.MouseButton.LeftButton, pos=ui_obj.tab_widget.tabBar().tabRect(1).center())
        # Check the result
        assert ui_obj.tab_widget.currentIndex() == 1
        assert isinstance(ui_obj.tab_widget.currentWidget(), QScrollArea)
        assert ui_obj.tab_widget.currentWidget().objectName() == 'settings_dialog_tab_editor_config'

        checkboxes = ui_obj.findChildren(QCheckBox)
        for checkbox in checkboxes:
            if isinstance(checkbox, QCheckBox):
                # Parse the object name in case it contains a combination of lexeme and setting keys
                _lexeme_key, setting_name = ui_obj.parse_object_name(checkbox.objectName())
                if setting_name == 'show_main_menu':
                    # Either Qt.CheckState.Checked or Qt.CheckState.Unchecked
                    prev_state = checkbox.checkState()
                    QTest.mouseClick(checkbox, Qt.MouseButton.LeftButton)
                    assert checkbox.checkState() != prev_state

    @pytest.mark.parametrize(
        "test_params_fixture, test_exp_params_fixture",
        [
            ('', ['', '']),
            (':', ['', '']),
            ('obj1_name', ['obj1_name', 'obj1_name']),
            ('obj1_name:setting1_name', ['obj1_name', 'setting1_name']),
            (':setting2_name', ['', 'setting2_name']),
        ],
        indirect=True
    )
    def test_parse_object_name(self, ui_obj: SettingsDialog, test_params_fixture, test_exp_params_fixture):
        """
        Covers:
        - SettingsDialog.parse_object_name
        """

        # Extract the test method name and expected parameters
        test_object_name = test_params_fixture
        exp_result = test_exp_params_fixture

        # Run the method under test
        result = ui_obj.parse_object_name(str(test_object_name))

        assert result == exp_result

    @pytest.mark.parametrize(
        "test_params_fixture, test_exp_params_fixture",
        [
            ('get_general_fields', 'settings_dialog_tab_general'),
            ('get_editor_config_fields', 'settings_dialog_tab_editor_config'),
            ('get_viewer_config_fields', 'settings_dialog_tab_viewer_config'),
            ('get_ai_config_fields', 'settings_dialog_tab_ai_config'),
        ],
        indirect=True
    )
    def test_get_tab_fields(self, mocker, ui_obj: SettingsDialog, test_params_fixture, test_exp_params_fixture):
        """
        Covers:
        - SettingsDialog.get_general_fields
        - SettingsDialog.get_editor_config_fields
        - SettingsDialog.get_viewer_config_fields
        - SettingsDialog.get_ai_config_fields
        """

        # Extract the test method name and expected parameters
        test_method_name = test_params_fixture
        exp_tab_name = test_exp_params_fixture

        # Mock the tab widget (QTabWidget instance)
        mock_tab_widget = MagicMock(spec=QTabWidget)
        mock_mock_tab_widget_add_tab = mocker.patch.object(mock_tab_widget, 'addTab')

        # Assign the mocked tab widget to the testing object
        ui_obj.tab_widget = mock_tab_widget

        # Mock QVBoxLayout's addWidget method
        mock_vbox_layout_add_widget = mocker.patch.object(QVBoxLayout, 'addWidget', return_value=None)

        # Mock the tab's main widget's setObjectName method
        mock_tab_widget_set_object_name = mocker.patch.object(QWidget, 'setObjectName')

        # Mock the scroll area containing the main tab widget
        mock_scroll_area_set_widget = mocker.patch.object(QScrollArea, 'setWidget')

        # Run the method under test
        test_method = getattr(ui_obj, str(test_method_name))
        assert callable(test_method), f"Method {test_method_name} is not callable."
        result = test_method()

        # Assert the result is of the expected type
        assert isinstance(result, list), "Expected the result to be a list."

        # Check that setObjectName methods were called with the expected parameters
        mock_tab_widget_set_object_name.assert_called_once_with(exp_tab_name)

        # Assert that the tab's main widget was added to the scroll area
        mock_scroll_area_set_widget.assert_called_once()
        assert 'QWidget' in str(mock_scroll_area_set_widget.call_args), "Expected a QWidget in the call arguments."

        # Verify that the tab was added to the tab widget
        mock_mock_tab_widget_add_tab.assert_called_once()

        # Validate the result by examining one of its dictionary values
        config_item = result.pop(0)
        assert 'callback' in config_item, "Expected 'callback' in the configuration item."
        assert callable(config_item['callback']), "Expected 'callback' to be callable."

        # Test the callback in the configuration
        mock_obj = MagicMock(spec=QObject)
        config_item['callback'](mock_obj)

        # Verify that the layout processed the callback and added the config's value to it
        mock_vbox_layout_add_widget.assert_called_once_with(mock_obj, alignment=Qt.AlignmentFlag.AlignTop)

    @pytest.mark.parametrize(
        "test_params_fixture, test_exp_params_fixture",
        [
            ({"type": HorizontalLineSpacer},
             (HorizontalLineSpacer, lambda v: True)),
            ({"type": QWidget, "size_policy": (QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)},
             (QWidget, lambda v: True)),
            ({"type": QCheckBox, "name": "obj1"},
             (QCheckBox, lambda v: v.objectName() == "obj1")),
            ({"type": QLineEdit, "placeholder_text": "text1"},
             (QLineEdit, lambda v: v.placeholderText() == "text1")),
            ({"type": QLabel, "props": {"setProperty": ("class", "class1")}},
             (QLabel, lambda v: v.property("class") == "class1")),
            ({"type": QSlider, "props": {"setTickInterval": 123}},
             (QSlider, lambda v: v.property("tickInterval") == 123)),
            ({"type": QPlainTextEdit, "text_lines": 3},
             (QPlainTextEdit, lambda v: v.height() > 1)),
            ({"type": QSpinBox, "props": {"setMinimum": 1, "setMaximum": 65536}},
             (QSpinBox, lambda v: (v.property("minimum") == 1, v.property("maximum") == 65536))),
            ({"type": QDoubleSpinBox, "props": {"setMinimum": 0.0, "setMaximum": 9.9}},
             (QDoubleSpinBox, lambda v: (v.property("minimum") == 0.0, v.property("maximum") == 9.9))),
            ({"type": QComboBox, "callback": lambda obj: obj.setProperty('param1', 'value1')},
             (QComboBox, lambda v: v.property('param1') == 'value1')),
            ({"type": EnumComboBox, "args": [
                sorted(SomethingEnum, key=lambda member: (not member.is_default, str(member.value)))]},
             (EnumComboBox, lambda v: (v.count() == 3, isinstance(v.currentIndex(), int)))),
            ({"type": LabelWithHint, "kwargs": {"tooltip": ("lexeme1_key", "lexeme1_text")}, "text": "text1"},
             (LabelWithHint, lambda v: (v.property("tooltip_lexeme") == "lexeme1_key", v.text == "text1"))),
            ({"type": DirPathLineEdit, "kwargs": {"default_directory": "dir1"}},
             (DirPathLineEdit, lambda v: (v.default_directory == "dir1"))),
            ({"type": FilePathLineEdit, "kwargs": {"ext_filter": "ext1"}},
             (FilePathLineEdit, lambda v: (v.ext_filter == "ext1"))),
        ],
        indirect=True
    )
    def test_create_setting_field(self, mocker, ui_obj: SettingsDialog, test_params_fixture, test_exp_params_fixture):
        # Extract test and expected params
        test_config = test_params_fixture
        exp_result_type, exp_check = test_exp_params_fixture

        # Run the method under test
        result_obj = ui_obj.create_setting_field(dict(test_config))

        # Assert the result is of the expected type
        assert isinstance(result_obj, exp_result_type)  # noqa

        # Test the callable check method
        assert callable(exp_check)
        assert exp_check(result_obj)
